// ignore_for_file:

@JS()
library channel;

import 'dart:async';
import 'dart:convert';
import 'dart:html' show Worker;

import 'package:built_value/serializer.dart';
import 'package:js/js.dart';

@anonymous
@JS()
abstract class MessageEvent {
  external String get data;
}

@JS('postMessage')
external void PostMessage(obj);

@JS('onmessage')
external set onMessage(f);

/// Returns a closure that produces a JSON String of the Typed message using the supplied serializer.
Function messageSerializer<T>(Serializers serializer) {
  JsonCodec jCdx = JsonCodec();
  return (T message) => jCdx.encode(serializer.serialize(message));
}

/// Returns a Typed Message from the JSON string using the supplied serializer.
Function messageDeserializer<T>(Serializers serializer) {
  JsonCodec jCdx = JsonCodec();
  return <T>(String jsonString) =>
      serializer.deserialize(jCdx.decode(jsonString));
}

/// Async stream for inbound and outbound messages.
class Channel<T> {
  final inbound = StreamController<T>();

  final outbound = StreamController<T>();

  static Serializers msgSerializer;
  static Type msgType;

  /// Returns a stream on which messages are listen for.
  Stream<T> get receivePort => inbound.stream;

  /// Returns a sink on which message are posted.
  Sink<T> get dispatchPort => outbound.sink;

  Type get supportedMessageType => msgType;

  /// Optionally pass in a BuildValue serializer when using build_value based messages.
  Channel({Serializers serializer}) {
    msgSerializer = serializer;
    msgType = T;
  }

  /// Returns a [Channel] that is Inter-connected with the passed in [Channel]
  Channel.interconnectWith(Channel<T> other) {
    other.outbound.stream.listen(inbound.add);
    outbound.stream.listen(other.inbound.add);
  }
}

/// Channel used to communicate with the passed in Worker.
class WorkerChannel<T> extends Channel<T> {
  final Worker _worker;
  static Function serializeMessage, deSerializeMessage;

  WorkerChannel(this._worker, {Serializers serializer}) {
    serializeMessage = messageSerializer<T>(serializer);
    deSerializeMessage = messageDeserializer(serializer);

    if (serializer == null) {
      outbound.stream.listen((msg) => _worker.postMessage(msg));
      _worker.onMessage.map((event) => event.data as T).listen(inbound.add);
    } else {
      outbound.stream
          .listen((typeMsg) => _worker.postMessage(serializeMessage(typeMsg)));
      _worker.onMessage
          .listen((event) => inbound.add(deSerializeMessage(event.data) as T));
    }
  }

  void terminateWorkerChannel() {
    _worker.terminate();
  }
}

/// Specialisation of channel suitable for use in the client running main client thread.
class WorkerChannelCounterPart<T> extends Channel<T> {
  static Function serializeMessage, deSerializeMessage;

  WorkerChannelCounterPart({Serializers serializer}) {
    serializeMessage = messageSerializer<T>(serializer);
    deSerializeMessage = messageDeserializer<T>(serializer);

    onMessage = allowInterop((event) {
      if (serializer == null) {
        inbound.add(event.data as T);
      } else {
        inbound.add(deSerializeMessage(event.data as String) as T);
      }
    });

    outbound.stream.listen((msg) {
      if (serializer == null) {
        PostMessage(msg);
      } else {
        PostMessage(serializeMessage(msg));
      }
    });
  }
}
